/*
  Jeremy Weatherford <xidus TA xidus TOD net>
  orbitclock.c

  No copyright -- steal this code!

  gcc -o orbitclock orbitclock.c -lX11 -L/usr/X11R6/lib

  v0.1 -- compiles on Irix, and is ANSIer C, thanks to jonh TA sgi TOD com
  v0.2 -- double-buffered, thanks to hugh4242 TA yahoo TOD com for pointer
          command line options for -geometry, -bg, -fg, -fg2, -rings, -date
	  rearrangement of rings thanks to lornix TA lornix TOD com
	  date option for including text in center
*/


#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <getopt.h>

#include <X11/Xlib.h>

#define min(x,y) (((x) < (y)) ? (x) : (y))
#define max(x,y) (((x) > (y)) ? (x) : (y))

/* default options */
int width=150, height=150, X=200, Y=200;
char *bg = "black", *fg = "orange", *fg2 = "blue";

int HOUR=0;
int MIN =1;
int SEC =2;
int drawdate = 0;
char *datestr = "%m/%d";
int datelines = 0;
char *fontname = "fixed";

/* command-line options */
static struct option long_opts[] = {
  {"help",      no_argument, 0, 'h'},
  {"geometry",  required_argument, 0, 'g'},
  {"bg",        required_argument, 0, 'b'},
  {"fg",        required_argument, 0, '1'},
  {"fg2",       required_argument, 0, '2'},
  {"rings",     required_argument, 0, 'r'},
  {"date",      optional_argument, 0, 'd'},
  {"font",      required_argument, 0, 'f'},
  {0, 0, 0, 0}
};
const char *optstring = "hg:b:1:2:r:d::f:";


void usage(void) {
  fprintf(stderr, "Usage: orbitclock [options]\n"
	  "\t-geometry WxH[+X+Y]     set window geometry\n"
	  "\t-bg <color name>        background color\n"
	  "\t-fg <color name>        foreground 1 color (tick)\n"
	  "\t-fg2 <color name>       foreground 2 color (block)\n"
	  "\t-rings <hms|smh|etc.>   ordering of rings from inside out (default hms)\n"
	  "\t-font <font>            font to use for date (implies -date)\n"
	  "\t-date[=optional format] draw date in center, strftime format string\n"
	  "\t                        literal '\\n' causes line break\n"
	  "\t                        example: -date=%%m/%%d/%%Y\\n%%H:%%M:%%S\n"
	  "\t                        equals sign is REQUIRED, don't ask me why.\n");
}



/* globals, yuck */
Display *dpy = NULL;
Window win;
GC gc, copygc;
XColor black, orange, blue;
XFontStruct *font;
Pixmap buffer;

/* time of last update */
time_t now = 0;

void draw(void) {
  const int circle = 360*64;
  const int center = min(width,height)/2;// - 10;
  int gap = -3*64*((200-min(center,150))/50);
  int offset = 90 * 64;
  int slice = center/3 - 10;
  int rank;
  int i;
  struct tm tm;
  int counts[3];

  now = time(NULL);
  localtime_r(&now, &tm);
  counts[HOUR] = (tm.tm_hour > 12) ? tm.tm_hour-12 : tm.tm_hour;
  counts[MIN] = tm.tm_min;
  counts[SEC] = tm.tm_sec;

  // clear to background
  XSetForeground(dpy, gc, black.pixel);
  XFillRectangle(dpy, buffer, gc, 0, 0, width, height);

  /* XFillArc doesn't let you give an inside radius, so we draw the
     outside circles first, always clearing the center to background */
  for (rank=2; rank >= 0; rank--) {
    int outer = center - ((2-rank)*slice);
    int inner = center-(slice-3) - ((2-rank)*slice);
    int count = (rank==HOUR) ? 12 : 60;
    int span = -circle / count;
    int group = (rank == HOUR) ? 3 : 5;
    XSetForeground(dpy, gc, orange.pixel);
    
    for(i=0; i<counts[rank]; i++) {
      int start = span*i+gap/2;
      int width;
      if (i+group <= counts[rank]) {
	width = span*group-gap/2;
	XSetForeground(dpy, gc, blue.pixel);
	i += group-1;
      } else {
	width = span-gap/2;
	XSetForeground(dpy, gc, orange.pixel);
      }
      XFillArc(dpy, buffer, gc, center-outer, center-outer,
	       outer*2, outer*2, offset+start, width);
    }
    XSetForeground(dpy, gc, black.pixel);
    XFillArc(dpy, buffer, gc, center-inner, center-inner,
	     inner*2, inner*2, 0, circle);
  }
  
  if (drawdate) {
    int h, w, y, hpad;
    char buf[512];
    int len;
    char *p, *q;

    strftime(buf, sizeof(buf), datestr, &tm);

    // font height
    h = font->max_bounds.ascent + font->max_bounds.descent;
    hpad = h + 2;

    p = buf;
    y = center - (hpad*datelines)/2 - h/2;
    // for each line
    do {
      q = strchr(p, '\n');
      if (q) len = q-p; else len = strlen(p);
      
      w = XTextWidth(font, p, len);

      XSetForeground(dpy, gc, orange.pixel);
      XDrawString(dpy, buffer, gc, center-w/2, y+font->ascent, p, len);

      y += hpad;

      p = q; if (p) p++;
    } while (p && q);
  }

  // update window from buffer
  XSetTile(dpy, copygc, buffer);
  XFillRectangle(dpy, win, copygc, 0, 0, width, height);
}

void expose(XExposeEvent *ev) {
  if (ev->count > 0) return;
  draw();
}

int errhandler(Display *dpy, XErrorEvent *ev) {
  char buf[512];

  XGetErrorText(dpy, ev->error_code, buf, sizeof(buf));
  printf("XError %d, %s\n", ev->error_code, buf);
  exit(1);
  return 1;
}

void alloc_color(char *color, XColor *ret) {
  XColor dummy;
  Colormap cmap = DefaultColormap(dpy, DefaultScreen(dpy));

  if (!XAllocNamedColor(dpy, cmap, color, ret, &dummy)) {
    fprintf(stderr, "Failed to allocate color '%s'\n", color);
    exit(1);
  }
}


int main(int argc, char *argv[]) {
  Window root;
  XGCValues gc_setup;
  int c, index;
  int *pcount[] = {&HOUR, &MIN, &SEC};
  char *p, *q;

  while ((c = getopt_long_only(argc, argv, optstring, long_opts, &index)) != -1) {
    switch (c) {
    case 'h':
      usage();
      exit(0);
      break;

    case 'g':
      XParseGeometry(optarg, &X, &Y, &width, &height);
      break;

    case 'b':
      bg = strdup(optarg);
      break;

    case '1':
      fg = strdup(optarg);
      break;

    case '2':
      fg2 = strdup(optarg);
      break;

    case 'r':
      for (c=0; c<3; c++) {
	p = strchr(optarg, "hms"[c]);
	if (!p) { usage(); exit(1); }
	*pcount[c] = (int)(p-optarg);
      }
      break;

    case 'd':
      if (optarg != NULL) {
	datestr = strdup(optarg);

	datelines = 0;
	/* convert literal "\\n" to '\n' */
	for(p=datestr, q=datestr; *p; ) {
	  if (*p == '\\' && p[1] == 'n') {
	    *q++ = '\n';
	    datelines++;
	    p += 2;;
	  } else
	    *q++ = *p++;
	}
	*q = '\0';
      }

      drawdate = 1;
      break;

    case 'f':
      fontname = strdup(optarg);
      drawdate = 1;
      break;

    case '?':
      usage();
      exit(0);
      break;
    }
  }

  dpy = XOpenDisplay(NULL);
  if (!dpy) {
    fprintf(stderr, "Failed to open default display.\n");
    exit(1);
  }

  XSetErrorHandler(errhandler);

  root = DefaultRootWindow(dpy);

  alloc_color(bg, &black);
  alloc_color(fg, &orange);
  alloc_color(fg2, &blue);

  win = XCreateSimpleWindow(dpy, root, X, Y, width, height, 0, 
			    black.pixel, black.pixel);
  XStoreName(dpy, win, "orbit clock");

  XSelectInput(dpy, win, ExposureMask | StructureNotifyMask);
  XMapWindow(dpy, win);

  buffer = XCreatePixmap(dpy, win, width, height, 
			 DefaultDepth(dpy, DefaultScreen(dpy)));


  gc = XCreateGC(dpy, win, 0, &gc_setup);
  XSetArcMode(dpy, gc, ArcPieSlice);

  copygc = XCreateGC(dpy, win, 0, &gc_setup);
  XSetFillStyle(dpy, copygc, FillTiled);

  font = XLoadQueryFont(dpy, fontname);
  if (!font) {
    fprintf(stderr, "Font not found, using fixed.\n");
    font = XLoadQueryFont(dpy, "fixed");
  }
  XSetFont(dpy, gc, font->fid);

  while (1) {
    struct timeval tv;
    XEvent ev;
    if (XPending(dpy)) {
      XNextEvent(dpy, &ev);
      switch(ev.type) {
      case Expose:
	expose((XExposeEvent *)&ev);
	break;
      case ConfigureNotify:
	width = ev.xconfigure.width;
	height = ev.xconfigure.height;
	XFreePixmap(dpy, buffer);
	buffer = XCreatePixmap(dpy, win, width, height, 
			       DefaultDepth(dpy, DefaultScreen(dpy)));
	draw();
	break;
      }
      continue;
    }
    if (now != time(NULL)) draw();
    tv.tv_usec = 10000; tv.tv_sec = 0;

    // wait a second...
    select(0, NULL, NULL, NULL, &tv);
  }
}
